import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt

__all__ = [
    "view_as_real",
    "view_as_complex",
    "view_as_real_mat",
    "view_as_complex_mat",
]


def view_as_complex(x: pxt.NDArray) -> pxt.NDArray:
    r"""
    View real-valued array as its complex-valued bijection.  (Inverse of :py:func:`~pyxu.util.view_as_real`.)

    Parameters
    ----------
    x: NDArray
        (..., 2N) real-valued array.

    Returns
    -------
    y: NDArray
        (..., N) complex-valued array.

    Examples
    --------

    .. code-block:: python3

       from pyxu.util import view_as_real, view_as_complex
       x = np.arange(6.0)      # array([0., 1., 2., 3., 4., 5.])
       y = view_as_complex(x)  # array([0.+1.j, 2.+3.j, 4.+5.j])
       view_as_real(y) == x    # True

    Notes
    -----
    Complex-valued inputs are returned unchanged.  For real-valued inputs, this function acts on the last axis as:

    .. math::

       y_n = x_{2n-1}+j \, x_{2n}, \qquad 1\leq n\leq N.

    See Also
    --------
    :py:func:`~pyxu.util.view_as_real`,
    :py:func:`~pyxu.util.view_as_real_mat`,
    :py:func:`~pyxu.util.view_as_complex_mat`
    """
    if _is_complex(x):
        return x

    try:
        r_dtype = x.dtype
        r_width = pxrt.Width(r_dtype)
        c_dtype = r_width.complex.value
    except Exception:
        raise ValueError(f"Unsupported dtype {r_dtype}.")
    assert x.shape[-1] % 2 == 0, "Last array dimension should be even-valued."

    c_sh = (*x.shape[:-1], x.shape[-1] // 2)
    y = x.view(c_dtype).reshape(c_sh)
    return y


def view_as_real(x: pxt.NDArray) -> pxt.NDArray:
    r"""
    View complex-valued array as its real-valued bijection.  (Inverse of :py:func:`~pyxu.util.view_as_complex`.)

    Parameters
    ----------
    x: NDArray
        (..., N) complex-valued array.

    Returns
    -------
    y: NDArray
        (..., 2N) real-valued array.

    Examples
    --------

    .. code-block:: python3

       from pyxu.util import view_as_real, view_as_complex
       x = np.r_[:3] + 1j * np.r_[2:5]   # array([0.+2.j, 1.+3.j, 2.+4.j])
       y = view_as_real(x)               # array([0., 2., 1., 3., 2., 4.])
       view_as_complex(y) == x           # True

    Notes
    -----
    Real-valued inputs are returned unchanged.  For complex-valued inputs, this function acts on the last axis as:

    .. math::

        y_{2n-1} = \mathcal{R}(x_n),
        \quad
        y_{2n} = \mathcal{I}(x_n),
        \quad 1\leq n\leq N,

    where :math:`\mathcal{R}, \mathcal{I}` denote the real/imaginary parts respectively.

    See Also
    --------
    :py:func:`~pyxu.util.view_as_complex`,
    :py:func:`~pyxu.util.view_as_real_mat`,
    :py:func:`~pyxu.util.view_as_complex_mat`
    """
    if _is_real(x):
        return x

    try:
        c_dtype = x.dtype
        c_width = pxrt.CWidth(c_dtype)
        r_dtype = c_width.real.value
    except Exception:
        raise ValueError(f"Unsupported dtype {c_dtype}.")

    r_sh = (*x.shape[:-1], 2 * x.shape[-1])
    y = x.view(r_dtype).reshape(r_sh)
    return y


def _is_real(x: pxt.NDArray) -> bool:
    try:
        return bool(pxrt.Width(x.dtype))
    except Exception:
        return False


def _is_complex(x: pxt.NDArray) -> bool:
    try:
        return bool(pxrt.CWidth(x.dtype))
    except Exception:
        return False


def view_as_real_mat(
    cmat: pxt.NDArray,
    real_input: bool = False,
    real_output: bool = False,
) -> pxt.NDArray:
    r"""
    View complex-valued matrix as its real-valued equivalent.  (Inverse of :py:func:`~pyxu.util.view_as_complex_mat`.)

    Useful to transform complex-valued matrix/vector products to their real-valued counterparts.

    Parameters
    ----------
    cmat: NDArray
        (M, N) complex-valued matrix.

    Returns
    -------
    rmat: NDArray
        The output shape depends on the values of `real_input` and `real_output`::

           | real_input | real_output | rmat.shape |
           |------------|-------------|------------|
           | False      | False       | (2M, 2N)   |
           | False      | True        | ( M, 2N)   |
           | True       | False       | (2M,  N)   |
           | True       | True        | ( M,  N)   |

    Examples
    --------

    .. code-block:: python3

       from pyxu.util import view_as_real_mat, view_as_complex_mat

       A = np.reshape(
           np.r_[:6] + 1j * np.r_[2:8], # array([[0.+2.j, 1.+3.j, 2.+4.j],
           newshape=(2, 3),             #        [3.+5.j, 4.+6.j, 5.+7.j]])
       )
       B = view_as_real_mat(A)          # array([[ 0., -2.,  1., -3.,  2., -4.],
                                        #        [ 2.,  0.,  3.,  1.,  4.,  2.],
                                        #        [ 3., -5.,  4., -6.,  5., -7.],
                                        #        [ 5.,  3.,  6.,  4.,  7.,  5.]])

    Notes
    -----
    * Real-valued matrices are returned unchanged.
    * Complex-valued matrices :math:`A\in\mathbb{C}^{M\times N}` are mapped into a real-valued matrix
      :math:`\hat{A}\in\mathbb{R}^{2M\times 2N}` defined, for :math:`1\leq n \leq N`, :math:`1\leq m\leq M` as

      .. math::

          \hat{A}_{2m-1,2n-1} = \mathcal{R}(A_{m,n}),
          & \quad
          \hat{A}_{2m-1,2n} = -\mathcal{I}(A_{m,n}),\\
          \hat{A}_{2m,2n-1} = \mathcal{I}(A_{m,n}),
          & \quad
          \hat{A}_{2m,2n}=\mathcal{R}(A_{m,n}).

      If ``real_[in|out]put=True``, then even columns/rows (or both) are furthermore dropped.  We have ``view_as_real(A
      @ x) = view_as_real_mat(A) @ view_as_real(x)``.

    See Also
    --------
    :py:func:`~pyxu.util.view_as_real`,
    :py:func:`~pyxu.util.view_as_complex`,
    :py:func:`~pyxu.util.view_as_complex_mat`
    """
    assert cmat.ndim == 2, f"cmat: expected a 2D array, got {cmat.ndim}-D."
    if _is_real(cmat):
        return cmat

    try:
        c_dtype = cmat.dtype
        c_width = pxrt.CWidth(c_dtype)
        r_dtype = c_width.real.value
    except Exception:
        raise ValueError(f"Unsupported dtype {c_dtype}.")

    xp = pxd.NDArrayInfo.from_obj(cmat).module()
    rmat = xp.zeros((2 * cmat.shape[0], 2 * cmat.shape[1]), dtype=r_dtype)
    rmat[::2, ::2], rmat[1::2, 1::2] = cmat.real, cmat.real
    rmat[::2, 1::2], rmat[1::2, ::2] = -cmat.imag, cmat.imag
    if real_input:
        rmat = rmat[:, ::2]
    if real_output:
        rmat = rmat[::2, :]
    return rmat


def view_as_complex_mat(
    rmat: pxt.NDArray,
    real_input: bool = False,
    real_output: bool = False,
) -> pxt.NDArray:
    r"""
    View real-valued matrix as its complex-valued equivalent.  (Inverse of :py:func:`~pyxu.util.view_as_real_mat`.)

    Useful to transform real-valued matrix/vector products to their complex-valued counterparts.

    Parameters
    ----------
    rmat: NDArray
        Real-valued matrix.  Accepted dimensions depend on the values of `real_input` and `real_output`::

           | real_input | real_output | rmat.shape |
           |------------|-------------|------------|
           | False      | False       | (2M, 2N)   |
           | False      | True        | ( M, 2N)   |
           | True       | False       | (2M,  N)   |
           | True       | True        | ( M,  N)   |

    Returns
    -------
    cmat: NDArray
        (M, N) complex-valued matrix.

    Examples
    --------

    .. code-block:: python3

       from pyxu.util import view_as_real_mat, view_as_complex_mat

       A = np.array([[ 0., -2.,  1., -3.,  2., -4.],
                     [ 2.,  0.,  3.,  1.,  4.,  2.],
                     [ 3., -5.,  4., -6.,  5., -7.],
                     [ 5.,  3.,  6.,  4.,  7.,  5.]])

       B = view_as_complex(A)  # array([[0.+2.j, 1.+3.j, 2.+4.j],
                               #        [3.+5.j, 4.+6.j, 5.+7.j]])

    Notes
    -----
    * Complex-valued matrices are returned unchanged.
    * Real-valued matrices are mapped into a complex-valued matrix :math:`{A}\in\mathbb{C}^{M\times N}` as follows:

      * :math:`\hat{A}\in\mathbb{R}^{2M\times 2N}`:  :math:`A_{m,n} = \hat{A}_{2m-1,2n-1} + j \hat{A}_{2m,2n-1}`,
      * :math:`\hat{A}\in\mathbb{R}^{M\times 2N}`:  :math:`A_{m,n} = \hat{A}_{m,2n-1} - j \hat{A}_{m,2n}`,
      * :math:`\hat{A}\in\mathbb{R}^{2M\times N}`:  :math:`A_{m,n} = \hat{A}_{2m-1,n} + j \hat{A}_{2m,n}`,
      * :math:`\hat{A}\in\mathbb{R}^{M\times N}`:  :math:`A_{m,n} = \hat{A}_{m,n} + 0j`,

      for :math:`1\leq n \leq N`, :math:`1\leq m\leq M`.

    See Also
    --------
    :py:func:`~pyxu.util.view_as_real`,
    :py:func:`~pyxu.util.view_as_complex`,
    :py:func:`~pyxu.util.view_as_real_mat`
    """
    assert rmat.ndim == 2, f"rmat: expected a 2D array, got {rmat.ndim}-D."
    if _is_complex(rmat):
        return rmat

    try:
        r_dtype = rmat.dtype
        r_width = pxrt.Width(r_dtype)
        c_dtype = r_width.complex.value
    except Exception:
        raise ValueError(f"Unsupported dtype {r_dtype}.")

    N_row, N_col = rmat.shape
    error_msg = lambda _: f"{_} array dimension should be even-valued."
    if real_input and real_output:
        cmat = rmat.astype(c_dtype)
    elif real_input and (not real_output):
        assert N_row % 2 == 0, error_msg("First")
        cmat = rmat[::2, :] + (1j * rmat[1::2, :])
    elif (not real_input) and real_output:
        assert N_col % 2 == 0, error_msg("Last")
        cmat = rmat[:, ::2] - (1j * rmat[:, 1::2])
    else:  # (not real_input) and (not real_output)
        assert N_row % 2 == 0, error_msg("First")
        assert N_col % 2 == 0, error_msg("Last")
        cmat = rmat[::2, ::2] + (1j * rmat[1::2, ::2])
    return cmat



import types
import typing as typ
import warnings

import numpy as np

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.operator.interop.source as px_src
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "IdentityOp",
    "NullOp",
    "NullFunc",
    "HomothetyOp",
    "DiagonalOp",
]


class IdentityOp(pxa.OrthProjOp):
    """
    Identity operator.
    """

    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        return pxu.read_only(arr)

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        return pxu.read_only(arr)

    def svdvals(self, **kwargs) -> pxt.NDArray:
        return pxa.UnitOp.svdvals(self, **kwargs)

    def asarray(self, **kwargs) -> pxt.NDArray:
        xp = kwargs.get("xp", pxd.NDArrayInfo.default().module())
        dtype = kwargs.get("dtype", pxrt.getPrecision().value)
        A = xp.eye(N=self.dim, dtype=dtype)
        return A

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        out = arr.copy()
        out /= 1 + damp
        return out

    def dagger(self, damp: pxt.Real, **kwargs) -> pxt.OpT:
        cst = 1 / (1 + damp)
        op = HomothetyOp(cst=cst, dim=self.dim)
        return op

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        return self.dim


class NullOp(pxa.LinOp):
    """
    Null operator.

    This operator maps any input vector on the null vector.
    """

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        self.lipschitz = 0

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.broadcast_to(
            xp.array(0, arr.dtype),
            (*arr.shape[:-1], self.codim),
        )

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.broadcast_to(
            xp.array(0, arr.dtype),
            (*arr.shape[:-1], self.dim),
        )

    def svdvals(self, **kwargs) -> pxt.NDArray:
        gpu = kwargs.get("gpu", False)
        xp = pxd.NDArrayInfo.from_flag(gpu).module()
        width = pxrt.getPrecision()

        D = xp.zeros(kwargs["k"], dtype=width.value)
        return D

    def gram(self) -> pxt.OpT:
        op = NullOp(shape=(self.dim, self.dim))
        return op.asop(pxa.SelfAdjointOp).squeeze()

    def cogram(self) -> pxt.OpT:
        op = NullOp(shape=(self.codim, self.codim))
        return op.asop(pxa.SelfAdjointOp).squeeze()

    def asarray(self, **kwargs) -> pxt.NDArray:
        xp = kwargs.get("xp", pxd.NDArrayInfo.default().module())
        dtype = kwargs.get("dtype", pxrt.getPrecision().value)
        A = xp.zeros(self.shape, dtype=dtype)
        return A

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        return 0


def NullFunc(dim: pxt.Integer) -> pxt.OpT:
    """
    Null functional.

    This functional maps any input vector on the null scalar.
    """
    op = NullOp(shape=(1, dim)).squeeze()
    op._name = "NullFunc"
    return op


def HomothetyOp(dim: pxt.Integer, cst: pxt.Real) -> pxt.OpT:
    """
    Constant scaling operator.

    Parameters
    ----------
    dim: Integer
        Dimension of the domain.
    cst: Real
        Scaling factor.

    Returns
    -------
    op: OpT
        (dim, dim) scaling operator.

    Note
    ----
    This operator is not defined in terms of :py:func:`~pyxu.operator.DiagonalOp` since it is array-backend-agnostic.
    """
    assert isinstance(cst, pxt.Real), f"cst: expected real, got {cst}."

    if np.isclose(cst, 0):
        op = NullOp(shape=(dim, dim))
    elif np.isclose(cst, 1):
        op = IdentityOp(dim=dim)
    else:  # build PosDef or SelfAdjointOp

        @pxrt.enforce_precision(i="arr")
        def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
            out = arr.copy()
            out *= _._cst
            return out

        def op_svdvals(_, **kwargs) -> pxt.NDArray:
            gpu = kwargs.get("gpu", False)
            xp = pxd.NDArrayInfo.from_flag(gpu).module()
            width = pxrt.getPrecision()

            D = xp.full(
                shape=kwargs["k"],
                fill_value=abs(_._cst),
                dtype=width.value,
            )
            return D

        @pxrt.enforce_precision(i=("arr", "damp"))
        def op_pinv(_, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
            out = arr.copy()
            scale = _._cst / (_._cst**2 + damp)
            out *= scale
            return out

        def op_dagger(_, damp: pxt.Real, **kwargs) -> pxt.OpT:
            scale = _._cst / (_._cst**2 + damp)
            op = HomothetyOp(cst=scale, dim=_.dim)
            return op

        def op_gram(_):
            return HomothetyOp(cst=_._cst**2, dim=_.dim)

        @pxrt.enforce_precision()
        def op_trace(_, **kwargs):
            out = _._cst * _.codim
            return out

        op = px_src.from_source(
            cls=pxa.PosDefOp if (cst > 0) else pxa.SelfAdjointOp,
            shape=(dim, dim),
            embed=dict(
                _name="HomothetyOp",
                _cst=cst,
            ),
            apply=op_apply,
            svdvals=op_svdvals,
            pinv=op_pinv,
            gram=op_gram,
            cogram=op_gram,
            trace=op_trace,
        )
        op.dagger = types.MethodType(op_dagger, op)
        op.lipschitz = abs(cst)
    return op.squeeze()


def DiagonalOp(
    vec: pxt.NDArray,
    enable_warnings: bool = True,
) -> pxt.OpT:
    r"""
    Element-wise scaling operator.

    Note
    ----
    :py:func:`~pyxu.operator.DiagonalOp` instances are **not arraymodule-agnostic**:
    they will only work with NDArrays belonging to the same array module as `vec`.  Moreover, inner computations may
    cast input arrays when the precision of `vec` does not match the user-requested precision.  If such a situation
    occurs, a warning is raised.

    Parameters
    ----------
    vec: NDArray
        (N,) diagonal scale factors.
    enable_warnings: bool
        If ``True``, emit a warning in case of precision mis-match issues.
    """
    assert len(vec) == np.prod(vec.shape), f"vec: {vec.shape} is not a DiagonalOp generator."
    if (dim := vec.size) == 1:  # Module-agnostic
        return HomothetyOp(cst=float(vec), dim=1)
    else:
        xp = pxu.get_array_module(vec)
        if pxu.compute(xp.allclose(vec, 0)):
            op = NullOp(shape=(dim, dim))
        elif pxu.compute(xp.allclose(vec, 1)):
            op = IdentityOp(dim=dim)
        else:  # build PosDef or SelfAdjointOp

            @pxrt.enforce_precision(i="arr")
            def op_apply(_, arr):
                if (_._vec.dtype != arr.dtype) and _._enable_warnings:
                    msg = "Computation may not be performed at the requested precision."
                    warnings.warn(msg, pxw.PrecisionWarning)
                out = arr.copy()
                out *= _._vec
                return out

            def op_asarray(_, **kwargs) -> pxt.NDArray:
                xp = kwargs.get("xp", pxd.NDArrayInfo.default().module())
                dtype = kwargs.get("dtype", pxrt.getPrecision().value)

                v = pxu.compute(_._vec.astype(dtype=dtype, copy=False))
                v = pxu.to_NUMPY(v)
                A = xp.diag(v)
                return A

            def op_gram(_):
                return DiagonalOp(
                    vec=_._vec**2,
                    enable_warnings=_._enable_warnings,
                )

            def op_svdvals(_, **kwargs):
                gpu = kwargs.get("gpu", False)
                xp = pxd.NDArrayInfo.from_flag(gpu).module()
                width = pxrt.getPrecision()

                k = kwargs["k"]
                which = kwargs.get("which", "LM")

                D = xp.abs(pxu.compute(_._vec))
                D = D[xp.argsort(D)]
                D = D.astype(width.value, copy=False)
                return D[:k] if (which == "SM") else D[-k:]

            @pxrt.enforce_precision(i=("arr", "damp"))
            def op_pinv(_, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    scale = _._vec / (_._vec**2 + damp)
                    scale[xp.isnan(scale)] = 0
                out = arr.copy()
                out *= scale
                return out

            def op_dagger(_, damp: pxt.Real, **kwargs) -> pxt.OpT:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    scale = _._vec / (_._vec**2 + damp)
                    scale[xp.isnan(scale)] = 0
                return DiagonalOp(
                    vec=scale,
                    enable_warnings=_._enable_warnings,
                )

            @pxrt.enforce_precision()
            def op_trace(_, **kwargs):
                return float(_._vec.sum())

            def op_estimate_lipschitz(_, **kwargs):
                # Calling LinOp's generic method=svd solver may fail since it relies on LinearOperator.
                # We insead use the fact that _lipschitz is computed exactly at construction time.
                return _._lipschitz

            op = px_src.from_source(
                cls=pxa.PosDefOp if pxu.compute(xp.all(vec > 0)) else pxa.SelfAdjointOp,
                shape=(dim, dim),
                embed=dict(
                    _name="DiagonalOp",
                    _vec=vec,
                    _enable_warnings=bool(enable_warnings),
                ),
                apply=op_apply,
                estimate_lipschitz=op_estimate_lipschitz,
                asarray=op_asarray,
                gram=op_gram,
                cogram=op_gram,
                svdvals=op_svdvals,
                pinv=op_pinv,
                trace=op_trace,
            )
            op.dagger = types.MethodType(op_dagger, op)
            op.lipschitz = float(abs(vec).max())
        return op.squeeze()


def _ExplicitLinOp(
    cls: pxt.OpC,
    mat: typ.Union[pxt.NDArray, pxt.SparseArray],
    enable_warnings: bool = True,
) -> pxt.OpT:
    r"""
    Build a linear operator from its matrix representation.

    Given a matrix :math:`\mathbf{A} \in \mathbb{R}^{N \times M}`, the *explicit linear operator* associated to
    :math:`\mathbf{A}` is defined as

    .. math::

       f_\mathbf{A}(\mathbf{x})
       =
       \mathbf{A}\mathbf{x},
       \qquad
       \forall \mathbf{x} \in \mathbb{R}^{M},

    with adjoint given by:

    .. math::

       f^{\ast}_{\mathbf{A}}(\mathbf{z})
       =
       \mathbf{A}^{T}\mathbf{z},
       \qquad
       \forall \mathbf{z} \in \mathbb{R}^{N}.

    Parameters
    ----------
    cls: OpC
        LinOp sub-class to instantiate.
    mat: NDArray, SparseArray
        (M, N) matrix generator.
        The input array can be *dense* or *sparse*.
        Accepted sparse arrays are:

        * CPU: COO/CSC/CSR/BSR/GCXS
        * GPU: COO/CSC/CSR
    enable_warnings: bool
        If ``True``, emit a warning in case of precision mis-match issues.

    Notes
    -----
    * :py:class:`~pyxu.operator.linop.base._ExplicitLinOp` instances are **not arraymodule-agnostic**:
      they will only work with NDArrays belonging to the same (dense) array module as `mat`.  Moreover, inner
      computations may cast input arrays when the precision of `mat` does not match the user-requested precision.  If
      such a situation occurs, a warning is raised.

    * The matrix provided to :py:func:`~pyxu.operator.linop.base._ExplicitLinOp` is used as-is and can be accessed via
      ``.mat``.
    """

    def _standard_form(A):
        fail_dense = False
        try:
            pxd.NDArrayInfo.from_obj(A)
        except Exception:
            fail_dense = True

        fail_sparse = False
        try:
            pxd.SparseArrayInfo.from_obj(A)
        except Exception:
            fail_sparse = True

        if fail_dense and fail_sparse:
            raise ValueError("mat: format could not be inferred.")
        else:
            return A

    def _matmat(A, b, warn: bool) -> pxt.NDArray:
        # A: (M, N) dense/sparse
        # b: (..., N) dense
        # out: (..., M) dense
        if (A.dtype != b.dtype) and warn:
            msg = "Computation may not be performed at the requested precision."
            warnings.warn(msg, pxw.PrecisionWarning)

        M, N = A.shape
        sh_out = (*b.shape[:-1], M)
        b = b.reshape((-1, N)).T  # (N, (...).prod)
        out = A.dot(b)  # (M, (...).prod)
        return out.T.reshape(sh_out)

    @pxrt.enforce_precision(i="arr")
    def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
        return _matmat(_.mat, arr, warn=_._enable_warnings)

    @pxrt.enforce_precision(i="arr")
    def op_adjoint(_, arr: pxt.NDArray) -> pxt.NDArray:
        return _matmat(_.mat.T, arr, warn=_._enable_warnings)

    def op_estimate_lipscthitz(_, **kwargs) -> pxt.Real:
        N = pxd.NDArrayInfo
        S = pxd.SparseArrayInfo

        try:  # Sparse arrays
            sdi = S.from_obj(_.mat)
            if sdi in (S.SCIPY_SPARSE, S.PYDATA_SPARSE):
                ndi = N.NUMPY
            else:  # S.CUPY_SPARSE
                ndi = N.CUPY
        except Exception:  # Dense arrays
            ndi = N.from_obj(_.mat)

        kwargs.update(xp=ndi.module())
        klass = _.__class__
        return klass.estimate_lipschitz(_, **kwargs)

    def op_asarray(_, **kwargs) -> pxt.NDArray:
        N = pxd.NDArrayInfo
        S = pxd.SparseArrayInfo
        xp = kwargs.get("xp", pxd.NDArrayInfo.default().module())
        dtype = kwargs.get("dtype", pxrt.getPrecision().value)

        try:  # Sparse arrays
            info = S.from_obj(_.mat)
            if info in (S.SCIPY_SPARSE, S.CUPY_SPARSE):
                f = lambda _: _.toarray()
            elif info == S.PYDATA_SPARSE:
                f = lambda _: _.todense()
            A = f(_.mat.astype(dtype))  # `copy` field not ubiquitous
        except Exception:  # Dense arrays
            info = N.from_obj(_.mat)
            A = pxu.compute(_.mat.astype(dtype, copy=False))

            if info == N.DASK:
                # Chunks may have been mapped to sparse arrays.
                # Call .asarray() again for a 2nd pass.
                _op = _ExplicitLinOp(_.__class__, mat=A)
                A = _op.asarray(xp=N.NUMPY.module(), dtype=A.dtype)
        finally:
            A = pxu.to_NUMPY(A)

        return xp.array(A, dtype=dtype)

    @pxrt.enforce_precision()
    def op_trace(_, **kwargs) -> pxt.Real:
        if _.dim != _.codim:
            raise NotImplementedError
        else:
            try:
                tr = _.mat.trace()
            except Exception:
                # .trace() missing for [PYDATA,CUPY]_SPARSE API.
                S = pxd.SparseArrayInfo
                info = S.from_obj(_.mat)
                if info == S.PYDATA_SPARSE:
                    # use `sparse.diagonal().sum()`, but array must be COO.
                    try:
                        A = _.mat.tocoo()  # GCXS inputs
                    except Exception:
                        A = _.mat  # COO inputs
                    finally:
                        tr = info.module().diagonal(A).sum()
                elif info == S.CUPY_SPARSE:
                    tr = _.mat.diagonal().sum()
                else:
                    raise ValueError(f"Unknown sparse format {_.mat}.")
            return float(tr)

    op = px_src.from_source(
        cls=cls,
        shape=mat.shape,
        embed=dict(
            _name="_ExplicitLinOp",
            mat=_standard_form(mat),
            _enable_warnings=bool(enable_warnings),
        ),
        apply=op_apply,
        adjoint=op_adjoint,
        asarray=op_asarray,
        trace=op_trace,
        estimate_lipschitz=op_estimate_lipscthitz,
    )
    return op



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class TestIdentityOp(conftest.OrthProjOpT):
    @pytest.fixture(params=[1, 10])
    def dim(self, request):
        return request.param

    @pytest.fixture
    def data_shape(self, dim):
        return (dim, dim)

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, dim, request):
        op = pxo.IdentityOp(dim=dim)
        return op, *request.param

    @pytest.fixture
    def data_apply(self, dim):
        arr = np.arange(dim)
        out = arr
        return dict(
            in_=dict(arr=arr),
            out=out,
        )



import itertools

import numpy as np
import pytest

import pyxu.info.ptype as pxt
import pyxu.opt.solver as pxsl
import pyxu_tests.opt.solver.conftest as conftest


class TestNLCG(conftest.SolverT):
    @staticmethod
    def spec_data(N: int) -> list[tuple[pxt.SolverC, dict, dict]]:
        klass = [
            pxsl.NLCG,
        ]

        funcs = conftest.funcs(N, seed=4)
        stream = conftest.generate_funcs(funcs, N_term=1)
        kwargs_init = [dict(f=f) for (f, *_) in stream]

        kwargs_fit = []
        param_sweep = dict(
            x0=[
                np.full((N,), 50),
                np.full((2, N), 50),  # multiple initial points
            ],
            variant=[
                "PR",
                "FR",
            ],
        )
        for config in itertools.product(*param_sweep.values()):
            d = dict(zip(param_sweep.keys(), config))
            kwargs_fit.append(d)

        data = itertools.product(klass, kwargs_init, kwargs_fit)
        return list(data)

    @pytest.fixture(params=spec_data(N=7))
    def spec(self, request) -> tuple[pxt.SolverC, dict, dict]:
        klass, kwargs_init, kwargs_fit = request.param
        return klass, kwargs_init, kwargs_fit

    @pytest.fixture
    def cost_function(self, kwargs_init) -> dict[str, pxt.OpT]:
        func = kwargs_init["f"]
        return dict(x=func)



import collections.abc as cabc
import inspect
import types
import typing as typ

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu


@pytest.fixture(params=pxd.supported_array_modules())
def xp(request) -> types.ModuleType:
    return request.param


def flaky(func: cabc.Callable, args: dict, condition: bool, reason: str):
    # XFAIL if func(**args) fails when `condition` satisfied.
    # This function is required when pytest.mark.xfail() cannot be used.
    # (Reason: `condition` only available inside test body.)
    try:
        out = func(**args)
        return out
    except Exception:
        if condition:
            pytest.xfail(reason)
        else:
            raise


def isclose(
    a: typ.Union[pxt.Real, pxt.NDArray],
    b: typ.Union[pxt.Real, pxt.NDArray],
    as_dtype: pxt.DType,
) -> pxt.NDArray:
    """
    Equivalent of `xp.isclose`, but where atol is automatically chosen based on `as_dtype`.

    This function always returns a computed array, i.e. NumPy/CuPy output.
    """
    atol = {
        np.dtype(np.half): 3e-2,  # former pxrt.Width.HALF
        pxrt.Width.SINGLE.value: 2e-4,
        pxrt.CWidth.SINGLE.value: 2e-4,
        pxrt.Width.DOUBLE.value: 1e-8,
        pxrt.CWidth.DOUBLE.value: 1e-8,
    }
    # Numbers obtained by:
    # * \sum_{k >= (p+1)//2} 2^{-k}, where p=<number of mantissa bits>; then
    # * round up value to 3 significant decimal digits.
    # N_mantissa = [10, 23, 52, 112] for [half, single, double, quad] respectively.

    prec = atol.get(as_dtype, pxrt.Width.DOUBLE.value)  # default only should occur for integer types
    eq = np.isclose(pxu.compute(a), pxu.compute(b), atol=prec)
    return eq


def allclose(
    a: pxt.NDArray,
    b: pxt.NDArray,
    as_dtype: pxt.DType,
) -> bool:
    """
    Equivalent of `all(isclose)`, but where atol is automatically chosen based on `as_dtype`.
    """
    return bool(np.all(isclose(a, b, as_dtype)))


def less_equal(
    a: pxt.NDArray,
    b: pxt.NDArray,
    as_dtype: pxt.DType,
) -> pxt.NDArray:
    """
    Equivalent of `a <= b`, but where equality tests are done at a chosen numerical precision.

    This function always returns a computed array, i.e. NumPy/CuPy output.
    """
    x = pxu.compute(a <= b)
    y = isclose(a, b, as_dtype)
    return x | y


class DisableTestMixin:
    """
    Disable certain tests based on user black-list.

    Example
    -------
    Consider the sample file below, where `test_1` and `test_3` should not be run.

    .. code-block:: python3

       # file: test_myclass.py

       import pytest
       import pyxu_tests.conftest as ct

       class TestMyClass(ct.DisableTestMixin):
           disable_test = {  # disable these tests
               "test_1",
               "test_3",
           }

           def test_1(self):
               self._skip_if_disabled()
               assert False  # should fail if not disabled

           def test_2(self):
               self._skip_if_disabled()
               assert True   # should pass

           def test_3(self):
               self._skip_if_disabled()
               assert False  # should fail if not disabled


    .. code-block:: bash

       $ pytest --quiet test_myclass.py

       s.s                           [100%]
       1 passed, 2 skipped in 0.59s             # -> test_[13]() were skipped/disabled.
    """

    disable_test: cabc.Set[str] = frozenset()

    def _skip_if_disabled(self):
        # Get name of function which invoked me.
        my_frame = inspect.currentframe()
        up_frame = inspect.getouterframes(my_frame)[1].frame
        up_finfo = inspect.getframeinfo(up_frame)
        up_fname = up_finfo.function
        if up_fname in self.disable_test:
            pytest.skip("disabled test")



import collections.abc as cabc
import contextlib
import enum
import functools
import numbers as nb

import numpy as np

import pyxu.info.ptype as pxt
import pyxu.util as pxu

__all__ = [
    "Width",
    "CWidth",
    "coerce",
    "enforce_precision",
    "EnforcePrecision",
    "getCoerceState",
    "getPrecision",
    "Precision",
]


@enum.unique
class Width(enum.Enum):
    """
    Machine-dependent floating-point types.
    """

    SINGLE = np.dtype(np.single)
    DOUBLE = np.dtype(np.double)

    def eps(self) -> pxt.Real:
        """
        Machine precision of a floating-point type.

        Returns the difference between 1 and the next smallest representable float larger than 1.
        """
        eps = np.finfo(self.value).eps
        return float(eps)

    @property
    def complex(self) -> "CWidth":
        """
        Returns precision-equivalent complex-valued type.
        """
        return CWidth[self.name]


@enum.unique
class CWidth(enum.Enum):
    """
    Machine-dependent complex-valued floating-point types.
    """

    SINGLE = np.dtype(np.csingle)
    DOUBLE = np.dtype(np.cdouble)

    @property
    def real(self) -> "Width":
        """
        Returns precision-equivalent real-valued type.
        """
        return Width[self.name]


class Precision(contextlib.AbstractContextManager):
    """
    Context Manager to locally redefine floating-point precision.

    Use this object via a with-block.

    Example
    -------
    .. code-block:: python3

       import pyxu.runtime as pxrt

       pxrt.getPrecision()                      # Width.DOUBLE
       with pxrt.Precision(pxrt.Width.SINGLE):
           pxrt.getPrecision()                  # Width.SINGLE
       pxrt.getPrecision()                      # Width.DOUBLE
    """

    def __init__(self, width: Width):
        self._width = width
        self._width_prev = getPrecision()

    def __enter__(self) -> "Precision":
        _setPrecision(self._width)
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        exc_raised = any(_ is not None for _ in [exc_type, exc_value, traceback])
        if exc_raised:
            pass

        _setPrecision(self._width_prev)
        return False if exc_raised else True


class EnforcePrecision(contextlib.AbstractContextManager):
    """
    Context Manager to locally disable effect of :py:func:`~pyxu.runtime.enforce_precision`.  [Default: enabled.]

    Use this object via a with-block.

    Example
    -------
    .. code-block:: python3

       import pyxu.runtime as pxrt

       pxrt.getCoerceState()                    # True
       with pxrt.EnforcePrecision(False):
           pxrt.getCoerceState()                # False
       pxrt.getCoerceState()                    # True
    """

    def __init__(self, state: bool):
        self._state = state
        self._state_prev = getCoerceState()

    def __enter__(self) -> "EnforcePrecision":
        _setCoerceState(self._state)
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        exc_raised = any(_ is not None for _ in [exc_type, exc_value, traceback])
        if exc_raised:
            pass

        _setCoerceState(self._state_prev)
        return False if exc_raised else True


def enforce_precision(
    i: pxt.VarName = frozenset(),
    o: bool = True,
    allow_None: bool = True,
) -> cabc.Callable:
    """
    Decorator to pre/post-process function parameters to enforce runtime FP-precision.

    Parameters
    ----------
    i: VarName
        Function parameters for which precision must be enforced to runtime's FP-precision.  Function parameter values
        must have a NumPy API, or be scalars.  None-valued parameters are allowed if `allow_None` is True (default).
    o: bool
        If True (default), ensure function's output (if any) has runtime's FP-precision.  If function's output does not
        have a NumPy API or is not scalar-valued, set `o` explicitly to False.
    allow_None: bool

    Example
    -------
    .. code-block:: python3

       import pyxu.runtime as pxrt

       @pxrt.enforce_precision(i='y', o=False)  # `i` can process multiple args: `i=('x','y')`.
       def f(x, y, z=1):
           print(x.dtype, y.dtype)
           return x + y + z

       x, y = np.arange(5), np.r_[0.5]
       print(x.dtype, y.dtype)  # int64, float64

       with pxrt.Precision(pxrt.Width.SINGLE):
           out = f(x,y)                         # (int64, float32) [printed inside f-call.]
       print(out.dtype)                         # float64 [would have been float32 if `o=True`.]
    """

    def decorator(func: cabc.Callable) -> cabc.Callable:
        @functools.wraps(func)
        def wrapper(*ARGS, **KWARGS):
            func_args = pxu.parse_params(func, *ARGS, **KWARGS)

            for k in [i] if isinstance(i, str) else i:
                if k not in func_args:
                    error_msg = f"Parameter[{k}] not part of {func.__qualname__}() parameter list."
                    raise ValueError(error_msg)
                elif func_args[k] is None:
                    if not allow_None:
                        raise ValueError(f"Parameter[{k}] cannot be None-valued.")
                else:
                    func_args[k] = coerce(func_args[k])

            out = func(**func_args)
            if o and (out is not None):
                out = coerce(out)
            return out

        return wrapper

    return decorator


def getPrecision() -> Width:
    """
    Query current FP precision.
    """
    state = globals()
    return state["__width"]


def getCoerceState() -> bool:
    """
    Query if :py:func:`~pyxu.runtime.coerce` is (currently) a no-op.
    """
    state = globals()
    return state["__coerce"]


def coerce(x):
    """
    Transform input to match runtime FP-precision.

    Parameters
    ----------
    x: Real, NDArray

    Returns
    -------
    y: Real, NDArray
        Input cast to the runtime FP-precision.  Fails if operation is impossible or unsafe. (I.e. casting
        complex-valued data.)

    Note
    ----
    This method is a NO-OP if :py:func:`~pyxu.runtime.getCoerceState()` returns ``False``.
    """
    if getCoerceState() is False:
        return x
    else:
        dtype = getPrecision().value
        try:
            if isinstance(x, pxt.Real):
                return np.array(x, dtype=dtype)[()]
            elif isinstance(x, nb.Number):
                raise  # other number categories cannot be converted.
            elif np.can_cast(x.dtype, dtype, casting="same_kind"):
                return x.astype(dtype, copy=False)  # cast warnings impossible
            else:
                raise
        except Exception:
            raise TypeError(f"Cannot coerce {type(x)} to scalar/array of precision {dtype}.")


def _setPrecision(width: Width):
    # For internal use only. It is recommended to modify FP-precision locally using the `Precision`
    # context manager.
    state = globals()
    state["__width"] = width


def _setCoerceState(s: bool):
    # For internal use only. It is recommended to modify coercion effect locally using the `Coerce`
    # context manager.
    state = globals()
    state["__coerce"] = s


__width = Width.DOUBLE  # default FP-precision.
__coerce = True  # default: coerce() activated.



from pyxu.info.plugin import _load_entry_points
from pyxu.opt.solver.adam import *
from pyxu.opt.solver.cg import *
from pyxu.opt.solver.nlcg import *
from pyxu.opt.solver.pds import *
from pyxu.opt.solver.pgd import *

_load_entry_points(globals(), group="pyxu.opt.solver")
